@@ -1,27 +1,34 @@
-from atomicwrites import atomic_write
+# Optional dependency: atomic_downloads
+try:
+    from atomicwrites import atomic_write
+
+    ATOMIC_DOWNLOADS_SUPPORT = True
+except ImportError:
+    ATOMIC_DOWNLOADS_SUPPORT = False
+
 
 from pathlib import Path
 import os
 
 from subwinder import utils
-from subwinder._ranking import rank_guess_media, rank_search_subtitles
+from subwinder._internal_utils import type_check
 from subwinder._request import request, Endpoints
-from subwinder.exceptions import (
-    SubAuthError,
-    SubDownloadError,
-    SubLangError,
-)
+from subwinder.exceptions import SubAuthError, SubDownloadError, SubLangError
 from subwinder.info import (
     build_media_info,
     Comment,
     EpisodeInfo,
     FullUserInfo,
+    GuessMediaResult,
+    MediaInfo,
     MovieInfo,
     SearchResult,
     ServerInfo,
+    SubtitlesInfo,
 )
 from subwinder.lang import lang_2s, lang_3s, lang_longs, LangFormat
 from subwinder.media import Media
+from subwinder.ranking import rank_guess_media, rank_search_subtitles
 
 
 def _build_search_query(query, lang):
@@ -177,25 +184,37 @@ class AuthSubwinder(Subwinder):
         self, downloads, download_dir=None, name_format="{upload_filename}"
     ):
         """
-        Attempts to download the `SearchResults` passed in as `downloads`. The download
+        Attempts to download the `SearchResult`s passed in as `downloads`. The download
         will attempt to place files in the same directory as the original file unless
         `download_dir` is provided. Files are automatically named according to the
         provided `name_format`.
         """
+        type_check(downloads, (list, tuple))
+
+        # Get all the subtitle file_ids from `downloads`
         # List of paths where the subtitle files should be saved
         download_paths = []
         for download in downloads:
-            media = download.media
-            subtitles = download.subtitles
+            # All downloads should be some container for `SubtitlesInfo`
+            type_check(download, (SearchResult, SubtitlesInfo))
+            if isinstance(download, SearchResult):
+                media = download.media
+                subtitles = download.subtitles
+            else:
+                subtitles = download
+                # There's no original media so fake it
+                media = MediaInfo.__new__(MediaInfo)
+                media.dirname = None
+                media.filename = None
 
             # Make sure there is enough context to save subtitles
-            if media.dirname is None and download_dir is None:
+            if media.get_dirname() is None and download_dir is None:
                 raise SubDownloadError(
                     "Insufficient context. Need to set either the `dirname` in"
                     f" {download} or `download_dir` in `download_subtitles`"
                 )
 
-            if media.filename is None:
+            if media.get_filename() is None:
                 # Hacky way to see if `media_name` is used in `name_format`
                 try:
                     _ = name_format.format(
@@ -216,13 +235,16 @@ class AuthSubwinder(Subwinder):
             # Store the subtitle file next to the original media unless `download_dir`
             # was set
             if download_dir is None:
-                dir_path = media.dirname
+                dir_path = media.get_dirname()
             else:
                 dir_path = Path(download_dir)
 
             # Format the `filename` according to the `name_format` passed in
-            media_name = None if media.filename is None else media.filename.stem
             upload_name = subtitles.filename.stem
+            if media.get_filename() is None:
+                media_name = None
+            else:
+                media_name = media.get_filename().stem
 
             filename = name_format.format(
                 media_name=media_name,
@@ -268,22 +290,36 @@ class AuthSubwinder(Subwinder):
             # Create the directories if needed, then save the file
             dirpath = fpath.parent
             dirpath.mkdir(exist_ok=True)
-            with atomic_write(fpath) as f:
-                f.write(subtitles)
 
-    def get_comments(self, search_results):
+            # Write atomically if possible, otherwise fall back to regular writing
+            if ATOMIC_DOWNLOADS_SUPPORT:
+                with atomic_write(fpath) as f:
+                    f.write(subtitles)
+            else:
+                with fpath.open("w") as f:
+                    f.write(subtitles)
+
+    def get_comments(self, sub_containers):
         """
         Get all `Comment`s for the provided `search_results` if there are any.
         """
-        subtitle_ids = [s.subtitles.id for s in search_results]
-        data = self._request(Endpoints.GET_COMMENTS, subtitle_ids)["data"]
+        type_check(sub_containers, (list, tuple))
+
+        ids = []
+        for sub_container in sub_containers:
+            type_check(sub_container, (SearchResult, SubtitlesInfo))
+
+            if isinstance(sub_container, SearchResult):
+                sub_container = sub_container.subtitles
+            ids.append(sub_container.id)
+        data = self._request(Endpoints.GET_COMMENTS, ids)["data"]
 
         # Group the results, if any, by the query order
-        groups = [[] for _ in search_results]
+        groups = [[] for _ in sub_containers]
         if data:
             for id, comments in data.items():
                 # Returned `id` has a leading _ for some reason so strip it
-                index = subtitle_ids.index(id[1:])
+                index = ids.index(id[1:])
                 groups[index] = data[id]
 
         # Pack results, if any, into `Comment` objects
@@ -310,38 +346,41 @@ class AuthSubwinder(Subwinder):
     def guess_media(
         self, queries, ranking_func=rank_guess_media, *rank_args, **rank_kwargs,
     ):
+        """
+        Same as `guess_media_unranked`, but selects the best result using `ranking_func`
+        """
+        results = self.guess_media_unranked(queries)
+
+        selected = []
+        for result, query in zip(results, queries):
+            selected.append(ranking_func(result, query, *rank_args, **rank_kwargs))
+
+        return selected
+
+    def guess_media_unranked(self, queries):
         """
         Attempts to guess the media described by each of the `queries`. A custom ranking
         function can be provided to better match the result with `ranking_func` where
         parameters can be passed to this function using `*args` and `**kwargs`.
         """
-        VALID_CLASSES = (list, tuple)
-        if not isinstance(queries, VALID_CLASSES):
-            raise ValueError(
-                "`guess_media` expects `queries` to be of type included in"
-                f" {VALID_CLASSES}, but got type '{type(queries)}'"
-            )
+        type_check(queries, (list, tuple))
 
         # Batch to 3 per api spec
-        return _batch(
-            self._guess_media, 3, [queries], ranking_func, *rank_args, **rank_kwargs,
-        )
+        return _batch(self._guess_media_unranked, 3, [queries])
 
-    def _guess_media(self, queries, ranking_func, *rank_args, **rank_kwargs):
+    def _guess_media_unranked(self, queries):
         data = self._request(Endpoints.GUESS_MOVIE_FROM_STRING, queries)["data"]
 
-        results = []
-        for query in queries:
-            result = ranking_func(data[query], query, *rank_args, **rank_kwargs)
+        # Special case: so `""` is just silently excluded from the response so force it
+        if "" in queries and "" not in data:
+            data[""] = {}
 
-            # Build the correct media info if anything was returned
-            results.append(None if result is None else build_media_info(result))
-
-        return results
+        # Convert the raw results to `GuessMediaResult`
+        return [GuessMediaResult.from_data(data[query]) for query in queries]
 
     # TODO: can we ensure that the `search_result` was matched using a file hash before
     #       calling this endpoint
-    def report_movie(self, search_result):
+    def report_media(self, sub_container):
         """
         Reports the subtitles tied to the `search_result`. This can only be done if the
         match was done using a `Media` object so that the subtitles can be tied to a
@@ -351,13 +390,41 @@ class AuthSubwinder(Subwinder):
         would be better to add a comment instead to give context as to why they're bad
         (`.add_comment("<meaningful-comment>", bad=True)`).
         """
-        self._request(
-            Endpoints.REPORT_WRONG_MOVIE_HASH, search_result.subtitles.sub_to_movie_id
-        )
+        type_check(sub_container, (SearchResult, SubtitlesInfo))
+
+        # Get subtitles file_id from `sub_container`
+        if isinstance(sub_container, SearchResult):
+            sub_container = sub_container.subtitles
+        sub_to_movie_id = sub_container.sub_to_movie_id
+
+        # This endpoint should only be used when the subtitles were matched by hash
+        if sub_to_movie_id is None:
+            raise ValueError(
+                "`report_media` can only be called if the subtitles were matched off a"
+                " search using a media files hash and size (This is done automatically"
+                " when searching with a `Media` object)."
+            )
+
+        self._request(Endpoints.REPORT_WRONG_MOVIE_HASH, sub_to_movie_id)
 
     def search_subtitles(
         self, queries, ranking_func=rank_search_subtitles, *rank_args, **rank_kwargs,
     ):
+        """
+        Same as `search_subtitles_unranked`, but the results are run through the
+        `ranking_func` first to try and determine the best result.
+        """
+        # Get all the results for the query
+        groups = self.search_subtitles_unranked(queries)
+
+        # And select the best results with the `ranking_func`
+        selected = []
+        for group, (query, _) in zip(groups, queries):
+            selected.append(ranking_func(group, query, *rank_args, **rank_kwargs))
+
+        return selected
+
+    def search_subtitles_unranked(self, queries):
         """
         Searches for any subtitles that match the provided `queries`. Queries are
         allowed to be `Media`, `MovieInfo`, or `EpisodeInfo` objects. A custom ranking
@@ -365,14 +432,23 @@ class AuthSubwinder(Subwinder):
         gets passed the provided `*args` and `**kwargs`.
         """
         # Verify that all the queries are correct before doing any requests
+        type_check(queries, (list, tuple, zip))
+
+        # Expand out the `zip` to a `list`
+        if isinstance(queries, zip):
+            queries = list(zip)
+
         VALID_CLASSES = (Media, MovieInfo, EpisodeInfo)
-        for query, lang_2 in queries:
-            if not isinstance(query, VALID_CLASSES):
+        for query_pair in queries:
+            if not isinstance(query_pair, (list, tuple)) or len(query_pair) != 2:
                 raise ValueError(
-                    "`search_subtitles` expects `queries` to contain objects"
-                    f" of {VALID_CLASSES}, but got type '{type(query)}'"
+                    "The `search_subtitles` variants expect a list of pairs of the form"
+                    "(<queryable>, <2 letter language code>)"
                 )
 
+            query, lang_2 = query_pair
+            type_check(query, VALID_CLASSES)
+
             if lang_2 not in lang_2s:
                 # Show both the 2-char and long name if invalid lang is given
                 lang_map = [f"{k} -> {v}" for k, v in zip(lang_2s, lang_longs)]
@@ -384,85 +460,97 @@ class AuthSubwinder(Subwinder):
 
         # This can return 500 items, but one query could return multiple so 20 is being
         # used in hope that there are plenty of results for each
-        return _batch(
-            self._search_subtitles,
-            20,
-            [queries],
-            ranking_func,
-            *rank_args,
-            **rank_kwargs,
-        )
+        return _batch(self._search_subtitles_unranked, 20, [queries],)
 
-    def _search_subtitles(self, queries, ranking_func, *rank_args, **rank_kwargs):
+    def _search_subtitles_unranked(self, queries):
         internal_queries = [_build_search_query(q, l) for q, l in queries]
         data = self._request(Endpoints.SEARCH_SUBTITLES, internal_queries)["data"]
 
         # Go through the results and organize them in the order of `queries`
         groups = [[] for _ in internal_queries]
-        for d in data:
-            groups[int(d["QueryNumber"])].append(d)
+        for raw_result in data:
+            # Results are returned in an arbitrary order so first figure out the query
+            query_index = int(raw_result["QueryNumber"])
+            query, _ = queries[query_index]
 
-        search_results = []
-        for group, (query, _) in zip(groups, queries):
-            result = ranking_func(group, query, *rank_args, **rank_kwargs)
+            # Go ahead and format the result as a `SearchResult`
+            result = SearchResult.from_data(raw_result)
+            result.media.set_dirname(query.get_dirname())
+            result.media.set_filename(query.get_filename())
 
-            # Get `SearchResult` setup if there is info for one
-            search_result = None
-            if result is not None:
-                search_result = SearchResult.from_data(result)
-                if isinstance(query, Media):
-                    # Media could have the original file information tied to it
-                    search_result.media.dirname = query.dirname
-                    search_result.media.filename = query.filename
+            groups[query_index].append(result)
 
-            search_results.append(search_result)
-
-        return search_results
+        return groups
 
     def suggest_media(self, query):
         """
         Suggest results for guesses of what media is described by `query`.
         """
+        type_check(query, str)
+
         data = self._request(Endpoints.SUGGEST_MOVIE, query)["data"]
 
         # `data` is an empty list if there were no results
         return [] if not data else [build_media_info(media) for media in data[query]]
 
-    def add_comment(self, search_result, comment_str, bad=False):
+    def add_comment(self, sub_container, comment_str, bad=False):
         """
         Adds the comment `comment_str` for the `search_result`. If desired you can
         denote that the comment is due to the result being `bad`.
         """
-        self._request(
-            Endpoints.ADD_COMMENT, search_result.subtitles.id, comment_str, bad
-        )
+        type_check(sub_container, (SearchResult, SubtitlesInfo))
 
-    def vote(self, search_result, score):
+        # Get the `SubtitlesInfo` from `SearchResult`, then get subtitle id
+        if isinstance(sub_container, SearchResult):
+            sub_id_obj = sub_container.subtitles
+        sub_id = sub_id_obj.id
+
+        self._request(Endpoints.ADD_COMMENT, sub_id, comment_str, bad)
+
+    def vote(self, sub_container, score):
         """
-        Votes for the `search_result` with a score of `score`.
+        Votes for the `sub_id_obj` with a score of `score`.
         """
+        type_check(sub_container, (SearchResult, SubtitlesInfo))
+        type_check(score, int)
+
         if score < 1 or score > 10:
             raise ValueError(f"Subtitle Vote must be between 1 and 10, given '{score}'")
 
-        self._request(Endpoints.SUBTITLES_VOTE, search_result.subtitles.id, score)
+        # Get the `SubtitlesInfo` from `SearchResult`, then get subtitle id
+        if isinstance(sub_container, SearchResult):
+            sub_id_obj = sub_container.subtitles
+        sub_id = sub_id_obj.id
+
+        self._request(Endpoints.SUBTITLES_VOTE, sub_id, score)
 
     def auto_update(self, program_name):
         """
         Returns information about `program_name` that is supposed to be useful for
         automatic updates. (Version info and download urls)
         """
+        type_check(program_name, str)
+
         # Not sure if I should return this information in a better format
         return self._request(Endpoints.AUTO_UPDATE, program_name)
 
-    def preview_subtitles(self, results):
+    def preview_subtitles(self, sub_containers):
         """
         Gets a preview for the subtitles represented by `results`. Useful for being able
         to see part of the subtitles without eating into your daily download limit.
         """
-        ids = [q.subtitles.file_id for q in results]
+        type_check(sub_containers, (list, tuple))
+
+        # Get the subtitles file_ids from `sub_containers`
+        file_ids = []
+        for sub_container in sub_containers:
+            type_check(sub_container, (SearchResult, SubtitlesInfo))
+            if isinstance(sub_container, SearchResult):
+                sub_container = sub_container.subtitles
+            file_ids.append(sub_container.file_id)
 
         # Batch to 20 per api spec
-        return _batch(self._preview_subtitles, 20, [ids])
+        return _batch(self._preview_subtitles, 20, [file_ids])
 
     def _preview_subtitles(self, ids):
         data = self._request(Endpoints.PREVIEW_SUBTITLES, ids)["data"]
